/* 
Project					: GARMEN
Author					: Himanshu Kaul, D.Phil.
Date Started			: November 1, 2016 
Last Update				: September 30, 2021
License					: Creative Commons Attribution Non-Commercial 4.0
Copyright © Himanshu Kaul 2022
*/

#include "header.h"
#include <time.h>

//		AGENT TYPE: DEFINE THE VARIOUS AGENTS USED IN THE MODEL
#define		pluripotent_stem_cell		0
#define		ectoderm					1
#define		mesendoderm					2
#define		endoderm					3
#define 	trophectoderm				4
#define		oct4_high					5


//		TRANSCRIPTION FACTORS (NO PARTUCULAR ORDER)
#define		SOX17		0
#define		T			1
#define		CDX2		2
#define		SOX2		3
#define		OCT4		4
#define		WNT3		5
#define		NOGGIN		6
#define		BMP4		7
#define		DKK			8


//		AGENT RADIUS
const double radius 		= 10.0;

//		AGENT CYCLE (DIVISION RATE EVERY 30 ITERATIONS)
const int cell_life[4] = {30, 30, 30, 30};

static double* non_random_list;
static int non_random_list_length = 0;


//		BOOLEAN
typedef unsigned char Boolean;
#define 	False					0
#define 	True					1

//		DEFINING THE GRN TERNARY VARIABLES
#define		OFF						0
#define 	LOW						1
#define		HIGH					2


/* PI */
#ifndef		PI						
#define 	PI						3.142857143
#endif


int last_agent_id = -1;


//			CODE STARTS HERE			//

//		RANDOM NUMBER GENERATOR
double rnd()
{
	if (non_random_list_length > 0) {
		double non_random = *(non_random_list);
		non_random_list_length --;
		non_random_list ++;
		return non_random;
	} else {	
		return (double) rand()/RAND_MAX;
	}
}

//		OUTPUT MESSAGE - The message each agent is capable of outputting, which enables agent interactions
//		Messages outputted include the type of agent, its information, and the bonds it has established with other agents
//	 	This FLAME function enables communication between agents - refer to www.flame.ac.uk for details regarding FLAME functions 
void output_information(int iteration)
{
	add_information_message(
				get_id(),
				get_type(),
				get_x(),
				get_y(),
				get_monolayer_bonds(),
				get_counter(),
				iteration);	
}

//This function returns the new coordinates of daughter agents based on the coordinates of the parent agent during proliferation
double generate_coordinate(double old_coordinate)
{
	double information		= 0;
	
	while(information==0)
		information	= rnd()*radius/10;
	
	if(information<=0.5)
		information = -information;
	
	return old_coordinate + information;		
}

//	This is the collision detection and correction algorithm that ensures agents do not violate each other's boundaries. For details refer to 
//	Kaul et al., 2015, Scientific Reports 
void intercellular_force(int iteration)
{
	set_counter(get_counter()+1);		//This is a secondary agent clock
	
	double x1 = get_x();
	double y1 = get_y();
	double x2, y2;
	
	
	int monolayer_bonds = 0;		//contacts between cells inside the colony 
	int edge_bonds		= 0;		//contacts between cells at colony edge
	
	
	set_force_x(0.0);
	set_force_y(0.0);
	
	information_message = get_first_information_message();
	while(information_message)
	{
		if(information_message->iteration == iteration)
		{
			x2 = information_message->x;
			y2 = information_message->y;
			
			
			double distance = sqrt(((x1-x2)*(x1-x2))+((y1-y2)*(y1-y2)));							
			
			if(distance != 0)
			{
				double check = distance - 20.0;			
				double force;

				if(check <= 10)						
				{
					monolayer_bonds ++;

					if(information_message->type==trophectoderm)
						edge_bonds++;
					
				}
						
				if(check > 0.0) 
					{
					force = 0; 
					} 
				else{
						force = 0.5;
					}
				
				force *= 0.2;

				set_force_x(get_force_x() + 
		           	force * check *((x2-x1)/distance));
					
				set_force_y(get_force_y() + 
		           	force * check * ((y2-y1)/distance));
			}
		}
		information_message = get_next_information_message(information_message);
	}

	set_monolayer_bonds(monolayer_bonds);
	set_edge_bonds(edge_bonds);
	

	x1 += get_force_x();
	y1 += get_force_y();
	

	if (x1 < 0) {
		x1 = 0;
	}
	
	if (y1 < 0) {
		y1 = 0;
	}
	
	if (x1 > FLAME_environment_variable_substrate_width) {
		x1 = FLAME_environment_variable_substrate_width;
	}
	
	if (y1 > FLAME_environment_variable_substrate_width) {
		y1 = FLAME_environment_variable_substrate_width;
	}
	
	set_x(x1); 
	set_y(y1);

}

//	This module takes signal concentration data generated via COMSOL. The data from COMSOL is saved as .txt file and contains 4 columns.
// 	Columns in the 12 h data file represent x (column 1) and y (column 2) coordinates of the finite elements and [BMP4] (column 3) and [NOGGIN] (column 4) within the elements.
// 	Columns in the 24 - 48 h data files represent x and y coordinates of the finite elements and [WNT3] and [DKK] within the elements 
int signal_sensing()
{
	double xb, yb;
	double xc, yc, xd, yd;
	
	
	xb = get_x()-465.0;
	yb = get_y()-465.0;
	
	
	double dis; 
	double dis_temp = 100.0;
	double s1 = 0.0;
	double s2 = 0.0;
	double s3 = 0.0;
	double s4 = 0.0;
	
	char *datafile = "no file yet";
	
	if(get_counter()==60)
		datafile	= "12hours_data.txt";		//For all scenarios except the BMP4- pattern, this has [BMP4] and [NOGGIN] data. Refer to the ReadMe file
	else if(get_counter()==90)
		datafile	= "24hours_data.txt";		//This has [WNT3] and [DKK] data. Refer to the ReadMe file
	else if(get_counter()==120)
		datafile	= "36hours_data.txt";		//This has [WNT3] and [DKK] data. Refer to the ReadMe file
	else if(get_counter()==150)
		datafile	= "48hours_data.txt";		//This has [WNT3] and [DKK] data. Refer to the ReadMe file
	
	//This text reading only initiates at the following iterations. Refer to KNOWN ISSUES in the ReadME file if you are simulating patterns with perturbations
	if(get_counter()==60 || get_counter()==90 || get_counter()==120 || get_counter()==150) {					
		
	
		
	
	#define data_TSV "%s\t%s\t%s\t%s\n"					
	
	char d1[100], d2[100], d3[100], d4[100];
	FILE *report;
	int nb = 4;
	char ** name = (char **) malloc (nb*sizeof(char *));
	double *conc = (double *) malloc (nb*sizeof(double));

	int i = 0;
	for (i=0;i<nb;i++)
	{
		*(name+i) = get_morphogen()[i].amount.array;
		*(conc+i) = *(get_morphogen()[i].amount.array);
	}

	//Read last line from report and Update Agent Memory
	if((report = fopen(datafile, "r"))==NULL)
	{
			printf("error opening Report\n");
			exit(0);
	}

	int nb_l = 0;
	char Line[150000];
	
	//get the number of lines in the report
	while(fgets(Line, sizeof(Line), report))
	{
	nb_l++;
	}

	rewind(report);
	
	//Read Report Data
	for(i=0;i<nb_l;i++)
		{
			fscanf(report,data_TSV,&d1,&d2,&d3,&d4);				
			double xcor, ycor; 
			double scalar_bmp, scalar_wnt3, scalar_nog, scalar_dkk;

			xcor = (double)atof( d1 );
			ycor = (double)atof( d2 );
			
			scalar_bmp = (double)atof( d3 );
			scalar_nog = (double)atof( d4 );
			scalar_wnt3 = (double)atof( d3 );		
			scalar_dkk = (double)atof( d4 );

			scalar_bmp = round(scalar_bmp*100)/100;
			scalar_nog = round(scalar_nog*100)/100;
			scalar_wnt3 = round(scalar_wnt3*100)/100;		
			scalar_dkk = round(scalar_dkk*100)/100;
			
			
			
			
			xc = xcor*1000/17.45;			// *1000/17.45 is to convert the non-dimensional coordinates in COMSOL back to dimensional coordinates;
			yc = ycor*1000/17.45;			
			
			
			
			dis = sqrt(((xb-xc)*(xb-xc))+((yb-yc)*(yb-yc)));
			if (dis < dis_temp){
				dis_temp = dis;
				xd = xc;
				yd = yc;
				
				s1 = scalar_bmp;
				s2 = scalar_nog;
				s3 = scalar_wnt3;
				s4 = scalar_dkk;
			
				if(get_counter()==60)
				{
					get_morphogen()[0].amount.array[0] = s1;			//At this time point, we assumed that only [BMP4] has reached steady state so WNT3 and DKK have low values
					get_morphogen()[1].amount.array[0] = s2;
					get_morphogen()[2].amount.array[0] = 0.0010;			
					get_morphogen()[3].amount.array[0] = 0.0010;	
					
					//WHEN SIMULATING BMP4- CONDITION, COMMENT OUT LINES 311 - 314 AND USE THE ONES BELOW INSTEAD (THAT ARE COMMENTED FOR NOW)
					//get_morphogen()[0].amount.array[0] = 0.0;	
					//get_morphogen()[1].amount.array[0] = 0.0;
					//get_morphogen()[2].amount.array[0] = s3;			
					//get_morphogen()[3].amount.array[0] = s4;
				}
				else
				{
					get_morphogen()[2].amount.array[0] = s3;			//At these time points, [BMP4] profile does not change. For details, refer to Chhabra et al., 2019, PLoS Biology 
					get_morphogen()[3].amount.array[0] = s4;
				}
			}
	}
	
		   
	/* Close the report file */
   fclose(report);
	}
	
	return 0;
}

//	This module regulates proliferation of agents
int proliferation()
{
	int type1 = get_type();
	int allowable_neighbours = 4;		//An agent can only divide if it has less than 4 neighbours
	double stochasticity = rnd();
	
	
	double fraction = cell_life[type1] / 4;
	double prob = stochasticity * fraction;
	int new_cycle = round(prob);							//Cell cycle for the daughter agents is randomly set. See Kaul et al., 2022, Stem Cell Reports for details 	
	int num_cell_neighbours		= get_monolayer_bonds();	//The number of 'monolayer bonds' inform how many neighbours an agent have in the simulation
	
	//If agents are at the edge, assume they have 2 more neighbours
	if(get_x()==0 || get_x()==FLAME_environment_variable_substrate_width)
		num_cell_neighbours += 2;
	
	if(get_y()==0 || get_y()==FLAME_environment_variable_substrate_width)
		num_cell_neighbours += 2;
	
	
	if(num_cell_neighbours<allowable_neighbours)
		set_cell_cycle(get_cell_cycle()+1);
	
	// Add a new hpsc agent
	if(get_cell_cycle()>cell_life[type1])
	{
		int new_type		= get_type();
		double new_x		= generate_coordinate(get_x());
		double new_y		= generate_coordinate(get_y());
		
		
		add_hpsc_agent(
					get_id()+1,
					new_type,
					get_nucleus(),
					get_morphogen(),
					new_x,
					new_y,
					0,
					0,
					0,					//Monolayer Bonds
					0, 					//Edge Bonds
					new_cycle,			
					get_counter(),
					FLAME_environment_variable_substrate_width); 
	}
	
	if(get_cell_cycle()>30)
		set_cell_cycle(0);
	
	//This keeps the colony circular by eliminating agents that are 'off the micro-pattern'
	if(get_counter()>0)
		if(sqrt(((get_x()-465)*(get_x()-465))+((get_y()-465)*(get_y()-465)))>465)
			return 1;
		
	return 0;	
}

//	This module regulates how agents differentiate into the various germ-layer states
int differentiation()
{
	
	int sox17	= get_nucleus()[SOX17].activation_status.array[0];
	int t		= get_nucleus()[T].activation_status.array[0];
	int cdx2	= get_nucleus()[CDX2].activation_status.array[0];
	int sox2	= get_nucleus()[SOX2].activation_status.array[0];
	int oct4	= get_nucleus()[OCT4].activation_status.array[0];
	

	//rnd() reflectes the probability of differentiation if all conditions are met
	if(oct4==HIGH && t==OFF && sox2==HIGH && sox17==OFF && cdx2==OFF && rnd()<0.15)
		set_type(pluripotent_stem_cell);
	
	
	if(oct4!=OFF && sox2!=HIGH && t==OFF && sox17==OFF && rnd()<0.15)
		set_type(oct4_high);
	
		
	if(sox2==HIGH && oct4==OFF && t==OFF && sox17==OFF && rnd()<0.15)
		set_type(ectoderm);
	
	
	if((t==HIGH || t==LOW) && cdx2==OFF && sox2==OFF && sox17!=HIGH && rnd()<0.15)
		set_type(mesendoderm);
	
	
	if(((sox17==LOW && t!=HIGH) || sox17==HIGH) && cdx2==OFF && rnd()<0.15)	
		set_type(endoderm);
	
	if((cdx2==LOW || cdx2==HIGH) && rnd()<0.15)	
		set_type(trophectoderm);	

	
	return 0;
}

//	This is the gene regulatory network module
int genetic_circuitry()
{
	int sox17	= get_nucleus()[SOX17].activation_status.array[0];
	int t		= get_nucleus()[T].activation_status.array[0];
	int cdx2	= get_nucleus()[CDX2].activation_status.array[0];
	int sox2	= get_nucleus()[SOX2].activation_status.array[0];
	int oct4	= get_nucleus()[OCT4].activation_status.array[0];
	int wnt3	= get_nucleus()[WNT3].activation_status.array[0];
	int bmp4	= get_nucleus()[BMP4].activation_status.array[0];
	
	//This information comes directly from the COMSOL data files that were stored in the agent memory in the function SIGNAL_SENSING above
	double con_bmp	= get_morphogen()[0].amount.array[0];
	double con_nog	= get_morphogen()[1].amount.array[0];
	double con_wnt	= get_morphogen()[2].amount.array[0];
	double con_dkk	= get_morphogen()[3].amount.array[0];
	
	//==========Genetic Changes due to Chemical Composition of Media==========
	//If concentration>0.3, high; if 0.0 then off; otherwise low
	if(con_bmp>=0.3)						
		get_nucleus()[BMP4].activation_status.array[0]	= HIGH;
	else if(con_bmp==0.0)						
		get_nucleus()[BMP4].activation_status.array[0]	= OFF;
	else
		get_nucleus()[BMP4].activation_status.array[0]	= LOW;
	
	if(con_nog>=0.3)						
		get_nucleus()[NOGGIN].activation_status.array[0]	= HIGH;
	else if(con_nog==0.0)						
		get_nucleus()[NOGGIN].activation_status.array[0]	= OFF;
	else 
		get_nucleus()[NOGGIN].activation_status.array[0]	= LOW;
	
	
	if(con_wnt>=0.3)
		get_nucleus()[WNT3].activation_status.array[0]	= HIGH;
	else if(con_wnt==0.0)						
		get_nucleus()[WNT3].activation_status.array[0]	= OFF;
	else 			
		get_nucleus()[WNT3].activation_status.array[0]	= LOW;
		
	if(con_dkk>=0.3)
		get_nucleus()[DKK].activation_status.array[0]	= HIGH; 
	else if(con_dkk==0.0)						
		get_nucleus()[DKK].activation_status.array[0]	= OFF;
	else 			
		get_nucleus()[DKK].activation_status.array[0]	= LOW;
	//========================================================================
	
	
	if(get_counter()>60){
		
	int init_wnt	= con_wnt*1000;
	
	
	//REGULATES OCT4 PRIMARILY-------------------------------------------------------------------------------------------------
	if((con_wnt>=1.2 && con_wnt<=con_dkk && bmp4!=OFF) || (bmp4==OFF && con_wnt<con_dkk && oct4==HIGH && sox2==HIGH))	
	{
		get_nucleus()[OCT4].activation_status.array[0]		= HIGH;	
		get_nucleus()[SOX2].activation_status.array[0]		= LOW; 
	}
	else if(con_bmp>con_nog && (con_wnt<=con_dkk || wnt3==OFF) && oct4==HIGH) {		
		get_nucleus()[OCT4].activation_status.array[0]		= LOW;
		get_nucleus()[SOX2].activation_status.array[0]		= OFF;
		}
	else if(cdx2==HIGH)
		get_nucleus()[OCT4].activation_status.array[0]		= OFF;
	//---------------------------------------------------------------------------------------------------
	
	
	//REGULATES SOX2 PRIMARILY-------------------------------------------------------------------------------------------------
	if(con_wnt<1.2 && con_wnt<=con_dkk && (init_wnt==1 || con_wnt>0.5) && con_bmp<=con_nog && (sox17==OFF && t==OFF) && bmp4!=OFF) {
		get_nucleus()[SOX2].activation_status.array[0]		= HIGH;
		get_nucleus()[T].activation_status.array[0]			= OFF;	
		get_nucleus()[OCT4].activation_status.array[0]		= OFF;		
	}
	
	else if((bmp4==OFF || con_bmp<con_nog) && wnt3==OFF) {	
		get_nucleus()[SOX2].activation_status.array[0]		= HIGH;
		get_nucleus()[OCT4].activation_status.array[0]		= OFF;
		get_nucleus()[T].activation_status.array[0]			= OFF;
	}
	else if (con_wnt>con_dkk && (con_bmp<con_nog || bmp4==OFF)) {			
		get_nucleus()[T].activation_status.array[0]			= LOW;
		get_nucleus()[OCT4].activation_status.array[0]		= LOW;
		get_nucleus()[SOX2].activation_status.array[0]		= OFF; 
		}
	else if(t!=OFF || sox17!=OFF || con_bmp>con_nog)
		get_nucleus()[SOX2].activation_status.array[0]		= OFF;
	//---------------------------------------------------------------------------------------------------
	
	
	//REGULATES TBXT AND SOX17 PRIMARILY-------------------------------------------------------------------------------------------------
	if((con_wnt>=0.3 && con_wnt>con_dkk) && oct4!=OFF && SOX17==OFF)	
	{
		get_nucleus()[T].activation_status.array[0]			= HIGH;
		get_nucleus()[SOX2].activation_status.array[0]		= OFF;
		get_nucleus()[OCT4].activation_status.array[0]		= LOW;
	}
	
	if((oct4==HIGH && con_bmp>=con_nog && wnt3==LOW && con_wnt>0.002) || (oct4!=HIGH && con_wnt<con_dkk && con_bmp>con_nog && t==HIGH)){
		get_nucleus()[T].activation_status.array[0]				= LOW;
		get_nucleus()[SOX17].activation_status.array[0]			= LOW;
		get_nucleus()[SOX2].activation_status.array[0]			= OFF;	
		}
	if(sox2==HIGH)
		get_nucleus()[T].activation_status.array[0]		= OFF;
	
	if((con_wnt>=0.3 && con_wnt>con_dkk) && oct4!=OFF && SOX17==OFF)	
	{
		get_nucleus()[T].activation_status.array[0]			= HIGH;
		get_nucleus()[SOX2].activation_status.array[0]		= OFF;
		get_nucleus()[OCT4].activation_status.array[0]		= LOW;
	}
	
	if((oct4==HIGH && con_bmp>=con_nog && wnt3==LOW && con_wnt>0.002) || (oct4!=HIGH && con_wnt<con_dkk && con_bmp>con_nog && t==HIGH)){
		get_nucleus()[T].activation_status.array[0]			= LOW;
		get_nucleus()[SOX17].activation_status.array[0]		= LOW;
		get_nucleus()[SOX2].activation_status.array[0]		= OFF;	
		}
	if(sox2==HIGH)
		get_nucleus()[T].activation_status.array[0]			= OFF;
	
	
	if(con_bmp>1.0 && con_wnt>=con_dkk && con_wnt<=0.3 && oct4!=HIGH && con_wnt>0.002)
	{	
		get_nucleus()[T].activation_status.array[0]			= OFF;
		get_nucleus()[SOX17].activation_status.array[0]		= HIGH;
		get_nucleus()[OCT4].activation_status.array[0]		= OFF;
	}
	else if(con_bmp>1.0 && con_wnt<=0.3 && oct4!=OFF &&  con_wnt>0.002) 
	{	
		get_nucleus()[T].activation_status.array[0]			= LOW;
		get_nucleus()[SOX17].activation_status.array[0]		= LOW;
	}
	else if(bmp4==OFF && (con_wnt>0.3 && con_wnt>con_dkk) && oct4!=HIGH && (t==LOW || t==HIGH)) {
		get_nucleus()[SOX17].activation_status.array[0]		= LOW;
		get_nucleus()[T].activation_status.array[0]			= LOW;}
	else if(oct4==OFF && con_bmp>con_nog && con_bmp>=1.1)
		get_nucleus()[SOX17].activation_status.array[0]		= LOW;
	//---------------------------------------------------------------------------------------------------
	
	
	//REGULATES CDX2 PRIMARILY-------------------------------------------------------------------------------------------------
	if((con_bmp>=con_nog && (con_wnt<=0.1 && con_wnt<con_dkk) && oct4!=HIGH && bmp4!=OFF && wnt3==LOW)) {
		get_nucleus()[CDX2].activation_status.array[0]	= HIGH;
		get_nucleus()[T].activation_status.array[0]		= OFF;
		
		}
	if(bmp4==OFF && (con_wnt<=0.3 && wnt3!=OFF)) {				
		get_nucleus()[CDX2].activation_status.array[0]	= HIGH;
		get_nucleus()[OCT4].activation_status.array[0]	= OFF;
		}
	else if(wnt3==OFF && con_bmp>con_nog && oct4!=OFF) {	
		get_nucleus()[CDX2].activation_status.array[0]	= HIGH;
		get_nucleus()[SOX2].activation_status.array[0]	= OFF;
		}
	if(oct4==HIGH)
		get_nucleus()[CDX2].activation_status.array[0]	= OFF;
	else if(oct4==OFF && con_bmp>con_nog && con_bmp<1.1)
		get_nucleus()[CDX2].activation_status.array[0]	= HIGH;
	//---------------------------------------------------------------------------------------------------
		
	}
	
	return 0;	
}

int output_information_0()
{
	output_information(0);
	
	return 0;
}

int output_information_1()
{
	output_information(1);
	
	return 0;
}

int intercellular_force_1()
{
	intercellular_force(1);
	
	return 0;
}
